/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.config;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.function.Function;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.geant.idpextension.oidc.profile.config.OIDCProfileConfiguration;
import org.opensaml.profile.context.ProfileRequestContext;

import com.nimbusds.oauth2.sdk.auth.ClientAuthenticationMethod;

import net.shibboleth.idp.profile.config.AbstractConditionalProfileConfiguration;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NotLive;
import net.shibboleth.utilities.java.support.annotation.constraint.Unmodifiable;
import net.shibboleth.utilities.java.support.logic.Constraint;
import net.shibboleth.utilities.java.support.logic.FunctionSupport;
import net.shibboleth.utilities.java.support.primitive.StringSupport;

/**
 * Base class for OIDC protocol configuration, containing configuration bit for setting client authentication methods.
 */
public abstract class AbstractOIDCClientAuthenticableProfileConfiguration
        extends AbstractConditionalProfileConfiguration implements OIDCProfileConfiguration {

    /** Enabled token endpoint authentication methods. */
    @Nonnull private Function<ProfileRequestContext,Collection<String>> tokenEndpointAuthMethodsLookupStrategy;

    /**
     * Constructor.
     *
     * @param profileId Unique profile identifier
     */
    protected AbstractOIDCClientAuthenticableProfileConfiguration(@Nonnull @NotEmpty final String profileId) {
        super(profileId);
        
        setTokenEndpointAuthMethods(
                List.of(
                        ClientAuthenticationMethod.CLIENT_SECRET_BASIC.toString(),
                        ClientAuthenticationMethod.CLIENT_SECRET_POST.toString(),
                        ClientAuthenticationMethod.CLIENT_SECRET_JWT.toString(),
                        ClientAuthenticationMethod.PRIVATE_KEY_JWT.toString()));
    }

    /**
     * Get the enabled token endpoint authentication methods.
     * 
     * @param profileRequestContext profile request context
     * 
     * @return enabled token endpoint authentication methods
     */
    @Nonnull @NonnullElements @NotLive @Unmodifiable public List<String> getTokenEndpointAuthMethods(
            @Nullable final ProfileRequestContext profileRequestContext) {
        
        final Collection<String> methods = tokenEndpointAuthMethodsLookupStrategy.apply(profileRequestContext);
        if (methods != null) {
            return List.copyOf(methods);
        }
        return Collections.emptyList();
    }

    /**
     * Set the enabled token endpoint authentication methods.
     * 
     * @param methods What to set.
     */
    public void setTokenEndpointAuthMethods(@Nonnull @NonnullElements final Collection<String> methods) {
        Constraint.isNotNull(methods, "Collection of methods cannot be null");

        if (methods != null) {
            tokenEndpointAuthMethodsLookupStrategy =
                    FunctionSupport.constant(List.copyOf(StringSupport.normalizeStringCollection(methods)));
        } else {
            tokenEndpointAuthMethodsLookupStrategy = FunctionSupport.constant(null);
        }
    }

    /**
     * Set a lookup strategy for the enabled token endpoint authentication methods.
     *
     * @param strategy  lookup strategy
     */
    public void setTokenEndpointAuthMethodsLookupStrategy(
            @Nonnull final Function<ProfileRequestContext,Collection<String>> strategy) {
        tokenEndpointAuthMethodsLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

}